#!/usr/bin/env python
# -*- coding: iso-8859-15 -*-

# sched.py (Praktikum, Arbeitsblatt 5)
# 
# Vorlesung Betriebssysteme
# Hans-Georg Eer, Hochschule Mnchen
# hans-georg.esser@hm.edu

proccount = 0;   # Anzahl der Prozesse im System

# Status-Konstanten
S_ACTIVE=1
S_READY=2
S_BLOCKED=3
S_DONE=0

def statusname (i):
  if i == 1: return "Active"
  elif i == 2: return "Ready"
  elif i == 3: return "Blockd"
  elif i == 0: return "Done"
  else: return "Error"

tasks = []     # Liste der Tasks
current = -1   # aktueller Prozess
trace = []     # Wann lief welcher Prozess?

runqueue = []  # Prozess-IDs, die bereit sind (Warteschlange)
blocked  = []  # Prozess-IDs, die blockiert sind

# Methoden zum Setzen und Abfragen von Prozesseigenschaften

def set_cputime (pid, cputime):   tasks[pid]["cputime"] = cputime
def set_iotime (pid, iotime):     tasks[pid]["iotime"] = iotime
def set_status (pid, status):     tasks[pid]["status"] = status
def set_endtime (pid, endtime):   tasks[pid]["endtime"] = endtime
def set_firstruntime (pid,t):     tasks[pid]["firstruntime"] = t
def get_starttime (pid):   return tasks[pid]["starttime"]
def get_endtime (pid):     return tasks[pid]["endtime"]
def get_cputime (pid):     return tasks[pid]["cputime"]
def get_iotime (pid):      return tasks[pid]["iotime"]
def get_status (pid):      return tasks[pid]["status"]
def get_firstruntime(pid): return tasks[pid]["firstruntime"]
def inc_cputime (pid):            set_cputime (pid, get_cputime(pid)+1)
def inc_iotime (pid):             set_iotime (pid, get_iotime(pid)+1)

def activate (pid):
  for t in tasks: 
    if t["status"] == S_ACTIVE: t["status"] = S_READY
  tasks[pid]["status"] = S_ACTIVE

def get_freepid():
  global proccount
  proccount+=1
  return proccount-1

def create_process (starttime, behavior):
  pid = get_freepid()
  task={}
  task["starttime"] = starttime
  task["behavior"] = behavior
  task["firstruntime"] = -1   # noch nie gelaufen
  tasks.append (task)
  set_cputime(pid,0)
  set_iotime(pid,0)
  set_status(pid,S_READY)
  if behavior[0]==0:
    # Prozess faengt mit I/O-Phase an
    tasks[pid]["behavior"] = behavior[1:]
    set_status(pid, S_BLOCKED)
  set_endtime(pid,-1)  # noch nicht beendet
  return pid

def futureprocesses(t):
  # gibt alle PIDs zurueck fuer Prozesse, die nach Zeit t starten
  global proccount
  fp = []
  for pid in range(0,proccount):
      if get_starttime(pid) > t: fp.append(pid)
  return fp

def schedule():
  # Implementation des FCFS-Schedulers (First Come, First Served)
  global current, tasks, runqueue, blocked, current, cputime
  
  # falls aktueller Prozess noch bereit: weitermachen (FCFS)
  if (current!=-1) and (get_status(current) == S_ACTIVE):
    return current

  # falls weder bereite noch blockierte Prozesse: Ende
  if (runqueue == []) and (blocked == []):
    # vielleicht kommt noch einer?
    if futureprocesses(cputime) == []: return -2

  # falls nicht: nehme ersten Prozess aus Runqueue
  if (runqueue != []):
    return runqueue[0]
  else:
    # falls alle blockiert: idlen!
    return -1  #  alle blockiert


def ps ():
  global tasks,proccount,cputime,runqueue,blocked
  print "PID | Sta | End | CPU | I/O | Status | Verhalten"
  for pid in range(0,proccount):
    if cputime >= get_starttime(pid)-1:
      task = tasks[pid]
      print "%3d | %3d | %3d | %3d | %3d | %6s |" % (pid, 
        get_starttime(pid), get_endtime(pid),
        get_cputime(pid), get_iotime(pid), 
        statusname( get_status(pid) ) ),
      print task["behavior"]                             
  print "Runqueue:",runqueue,"  Blocked:",blocked
  print
  return

def init ():
  from sys import argv  # Argumente auswerten; Name der Konf.-Datei
  from sys import exit
  try:
    filename = argv[1]
    f = open (filename, "r")
    lines = f.readlines()
    f.close ()
  except:
    print "Fehler: Bitte Dateinamen der Konfigurationsdatei angeben."
    exit()
  
  for l in lines:
    if l == "\n": return
    (starttime,times) = l[:-1].split(":")
    starttime = int(starttime)
    times = times.split(",")
    behavior=[]
    for t in times:
      behavior.append(int(t))
    # print "Startzeit:",starttime, "Verhalten:", behavior
    create_process (starttime, behavior)
  return


def update_blocked_processes():
  global current, cputime, runqueue, blocked
  for pid in blocked:
    if cputime >= get_starttime(pid):
      tasks[pid]["behavior"][0] -= 1   # verbleibende Wartezeit reduzieren
      inc_iotime(pid)
      if tasks[pid]["behavior"][0] == 0:
        # Aktuelle I/O-Phase abgeschlossen
        tasks[pid]["behavior"] = tasks[pid]["behavior"][1:]
        set_status(pid,S_READY)
        blocked.remove(pid)
        if tasks[pid]["behavior"][0] == -1:
          set_status(pid,S_DONE)
          set_endtime(pid,cputime)
        else:
          runqueue.append(pid)
  return

def run_current():
  global current, cputime, runqueue, blocked
  if get_firstruntime(current) == -1:
    # falls Prozess zum ersten Mal laeuft: Zeit merken
    set_firstruntime(current, cputime)
  tasks[current]["behavior"][0] -= 1   # verbleibende Rechenzeit reduzieren
  inc_cputime(current)
  if tasks[current]["behavior"][0] == 0:
    tasks[current]["behavior"] = tasks[current]["behavior"][1:]
    set_status(current,S_BLOCKED)
    runqueue.remove(current)
    if tasks[current]["behavior"][0] == -1:
      set_status(current,S_DONE)
      set_endtime(current,cputime)
      if current in runqueue: runqueue.remove(current)
    else:
      blocked.append(current)
  return


def log_to_trace(status):
  global trace
  if status == -1: status = "NONE"
  elif status == -2: status = "END"
  trace.append(status)


def stats():
  global cputime
  print
  print "Endzeit: %d" % cputime
  print "Trace:",trace

  print "Laufzeiten:"
  print "PID Ankunft Rechenz Startze Endzeit TurnAro  Quotient"
  print "-----------------------------------------------------"

  for pid in range(0,proccount):
    stime  = get_starttime(pid)
    etime  = get_endtime(pid)+1
    frtime = get_firstruntime(pid)
    cputime = get_cputime(pid)
    if cputime != 0:
      ratio = (etime-stime+0.0)/cputime
    else:
      ratio = -1
    print "%3d %7d %7d %7d %7d %7d %9.4f" % (pid, 
      stime, cputime, frtime, etime, etime-stime, ratio)

init ()
print "Anzahl tasks: ", proccount


cputime = 0   # Init. der Uhr
finished = 0


ps()
# Hauptschleife
while not finished:

  # neue Prozesse angekommen?
  for pid in range(0,proccount):
    if get_starttime(pid) == cputime:
      if get_status(pid)==S_BLOCKED:
        blocked.append(pid)
      else:
        runqueue.append(pid)

  current = schedule()
  log_to_trace (current)
  if current >= 0:
    print "Zeit: %4d, Es laeuft PID: %3d [%3d]" % (cputime,current,get_cputime(current))
    run_current()  # aktuellen Prozess ausfuehren
    # bei allen blockierten Prozesse Wartezeit runterzaehlen
    update_blocked_processes()
  elif current == -1:
    print "Zeit: %4d, alle blockiert" % cputime
    # bei allen blockierten Prozesse Wartezeit runterzaehlen
    update_blocked_processes()
  else:  
    finished = 1
    break
  ps()
  cputime += 1

stats()
